//
// Lighthouse3D.com
//
//
// This demo was built for learning purposes only.
// Some code could be severely optimised, but I tried to
// keep as simple and clear as possible.
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//
// If you do use it I would love to hear about it. Just post a comment
// at Lighthouse3D.com

// Have Fun :-)

//
// Program for lecturing purposes by Marek Vinkler (xvinkl@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//


#include <cmath>

// Include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

// include DevIL for image loading
#include <IL/il.h>

#include "processing.h"

// Use Very Simple Libs
#include "vsMathLib.h"
#include "vsShaderLib.h"
#include "vsLogLib.h"

//----------------------------------------------------------------
// TODO - _NOEFFECT:
// 1) Inspect the functions changeSize, renderScene, setupShaders and init.
// 2) Do the tasks in pass.vert.
// 3) Inspect noeffect.frag.

// TODO - _GRAYSCALE:
// 1) Nothing to do :-)

// TODO - _NEGATIVE:
// 1) Nothing to do :-)

// TODO - _THRESHOLD:
// 1) Set the uniform variable for threshold to 0.45f.

// TODO - _BLUR:
// 1) Nothing to do :-)

// TODO - _SHARPEN:
// 1) Inspect the init function to see how the convolution kernel is set.
// 2) See how awkward OpenGL is with non-graphics data.

// TODO - _EDGEDETECT:
// 1) Inspect the init function to see how the convolution kernel is set.
//----------------------------------------------------------------

#define _NOEFFECT		0
#define _GRAYSCALE		1
#define _NEGATIVE		2
#define _THRESHOLD		3
#define _BLUR			4	
#define _SHARPEN		5
#define _EDGEDETECT		6

#define _EFFECT _NOEFFECT
//----------------------------------------------------------------

VSMathLib *vsml;
VSShaderLib shader;

VSLogLib mLogError, mLogInfo;

// Vao
GLuint vao;

// Texture
unsigned int texID, kernelID;

unsigned int loadRGBATexture(std::string filename, bool mipmap = true, bool compress = false, GLenum aFilter = GL_LINEAR, GLenum aRepMode = GL_REPEAT);


// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h)
{
	float ratio;

	// Prevent a divide by zero, when window is too short
	if(h == 0)
		h = 1;

	// Set the viewport to be the entire window
	glViewport(0, 0, w, h);

	ratio = (1.0f * w) / h;
	vsml->loadIdentity(VSMathLib::PROJECTION);
#if 0
	vsml->perspective(53.13f, ratio, 0.1f, 10000.0f);
#else
	vsml->ortho(-1.f, 1.f, -1.f, 1.f, 0.1f, 10000.0f);
#endif
}


// ------------------------------------------------------------
//
// Render stuff
//

void renderScene(void)
{
	glClear(GL_COLOR_BUFFER_BIT);

	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::MODEL);

	// Set camera
	vsml->lookAt(0.f, 0.f, 2.f, 0,0,0, 0,1,0);

	// Use our shader
	glUseProgram(shader.getProgramIndex());

	// Send matrices to uniform buffer
	vsml->matricesToGL();

	// Render VAO
	glBindVertexArray(vao);
	glDrawElements(GL_TRIANGLES, faceCount*3, GL_UNSIGNED_INT, 0);

	// Swap buffers
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;
	}

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// --------------------------------------------------------
//
// Shader Stuff
//

bool setupShaders()
{
	// Shader for models
	shader.init();
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pass.vert");

#if (_EFFECT == _NOEFFECT)
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/noeffect.frag");
#elif (_EFFECT == _GRAYSCALE)
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/grayscale.frag");
#elif (_EFFECT == _NEGATIVE)
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/negative.frag");
#elif (_EFFECT == _THRESHOLD)
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/threshold.frag");
#elif (_EFFECT == _BLUR)
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/blur.frag");
#elif (_EFFECT == _SHARPEN)
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/convolve.frag");
#elif (_EFFECT == _EDGEDETECT)
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/convolve.frag");
#endif

	// Set semantics for the shader variables
	shader.setProgramOutput(0, "outputF");
	shader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");

	printf("InfoLog\n%s\n\n", shader.getAllInfoLogs().c_str());
	shader.prepareProgram();

	shader.setUniform("texUnit", 0); // Add sampler uniform

	return(shader.isProgramLinked());
}


// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//

// helper function for derived classes
// loads an image and defines an 8-bit RGBA texture
unsigned int loadRGBATexture(std::string filename, bool mipmap, bool compress, GLenum aFilter, GLenum aRepMode)
{
	ILboolean success;
	unsigned int imageID;
	GLuint textureID = 0;

	// Load Texture Map
	ilGenImages(1, &imageID); 
	
	ilBindImage(imageID); /* Binding of DevIL image name */
	ilEnable(IL_ORIGIN_SET);
	ilOriginFunc(IL_ORIGIN_LOWER_LEFT); 
	success = ilLoadImage((ILstring)filename.c_str());

	if (!success) {
		VSLOG(mLogError, "Couldn't load texture: %s", 
							filename.c_str());
		// The operation was not sucessfull 
		// hence free image and texture 
		ilDeleteImages(1, &imageID); 
		return 0;
	}

	// Add information to the log

	VSLOG(mLogInfo, "Texture Loaded: %s", filename.c_str());
	printf("Width: %d, Height %d, Bytes per Pixel %d", 
				ilGetInteger(IL_IMAGE_WIDTH),
				ilGetInteger(IL_IMAGE_HEIGHT),
				ilGetInteger(IL_IMAGE_BYTES_PER_PIXEL));

	std::string s;
	switch(ilGetInteger(IL_IMAGE_FORMAT)) {
		case IL_COLOR_INDEX      : s =  "IL_COLOR_INDEX"; break;     
		case IL_ALPHA		     : s =  "IL_ALPHA"; break;	
		case IL_RGB              : s =  "IL_RGB"; break;     
		case IL_RGBA             : s =  "IL_RGBA"; break;     
		case IL_BGR              : s =  "IL_BGR"; break;     
		case IL_BGRA             : s =  "IL_BGRA"; break;     
		case IL_LUMINANCE        : s =  "IL_LUMINANCE"; break;     
		case IL_LUMINANCE_ALPHA  : s =  "IL_LUMINANCE_ALPHA"; break;
	}
	printf(" Format %s", s.c_str());
	
	switch(ilGetInteger(IL_IMAGE_TYPE)) {
		case IL_BYTE           : s =  "IL_BYTE"; break;     
		case IL_UNSIGNED_BYTE  : s =  "IL_UNSIGNED_BYTE"; break;	
		case IL_SHORT          : s =  "IL_SHORT"; break;     
		case IL_UNSIGNED_SHORT : s =  "IL_UNSIGNED_SHORT"; break;     
		case IL_INT            : s =  "IL_INT"; break;     
		case IL_UNSIGNED_INT   : s =  "IL_UNSIGNED_INT"; break;     
		case IL_FLOAT          : s =  "IL_FLOAT"; break;     
		case IL_DOUBLE         : s =  "IL_DOUBLE"; break;
		case IL_HALF           : s =  "IL_HALF"; break;
	}
	printf(" Data type:  %s\n", s.c_str());

	/* Convert image to RGBA */
	ilConvertImage(IL_RGBA, IL_UNSIGNED_BYTE); 
	// Set filters
	GLenum minFilter = aFilter;
	if (aFilter == GL_LINEAR && mipmap) {
		minFilter = GL_LINEAR_MIPMAP_LINEAR;
	}
	else if (aFilter == GL_NEAREST && mipmap){
		minFilter = GL_NEAREST_MIPMAP_LINEAR;
	}
	GLenum type;
	if (compress)
		type = GL_RGBA;
	else
		type = GL_COMPRESSED_RGBA;


	/* Create and load textures to OpenGL */
	glGenTextures(1, &textureID); /* Texture name generation */
	glBindTexture(GL_TEXTURE_2D, textureID); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, aFilter); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, aRepMode);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, aRepMode);
	glTexImage2D(GL_TEXTURE_2D, 0, type, 
					ilGetInteger(IL_IMAGE_WIDTH),
					ilGetInteger(IL_IMAGE_HEIGHT), 
					0, GL_RGBA, GL_UNSIGNED_BYTE,
					ilGetData()); 

	// Mipmapping?
	if (mipmap)
		glGenerateMipmap(GL_TEXTURE_2D);

	glBindTexture(GL_TEXTURE_2D,0);

	/* Because we have already copied image data into texture data
	we can release memory used by image. */
	ilDeleteImages(1, &imageID); 

	return textureID;
}

unsigned int setKernelTexture(float* coefs, int width, int height)
{
	/* Create and load textures to OpenGL */
	GLuint textureID = 0;
	glGenTextures(1, &textureID); /* Texture name generation */
	glBindTexture(GL_TEXTURE_2D, textureID); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_R32F, 
					width,
					height, 
					0, GL_RED, GL_FLOAT,
					coefs);
	glBindTexture(GL_TEXTURE_2D, 0);

	return textureID;
}

bool init()
{
	// Create the VAO
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	// Create buffers for our vertex data
	GLuint buffers[2];
	glGenBuffers(2, buffers);

	// Vertex coordinates buffer
	glBindBuffer(GL_ARRAY_BUFFER, buffers[0]);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);
	glEnableVertexAttribArray(VSShaderLib::VERTEX_COORD_ATTRIB);
	glVertexAttribPointer(VSShaderLib::VERTEX_COORD_ATTRIB, 4, GL_FLOAT, 0, 0, 0);

	// Index buffer
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffers[1]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(faceIndex), faceIndex, GL_STATIC_DRAW);

	// Unbind the VAO
	glBindVertexArray(0);

	/* Initialization of DevIL */
	ilInit();

	// Texture setup
	texID = loadRGBATexture("textures/lena_std.tif", false);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texID);

#if (_EFFECT == _THRESHOLD)
	shader.setUniform("threshold", 0.45f); // Must be here, after shader.prepareProgram(); that links the program!!!
#elif (_EFFECT == _SHARPEN) || (_EFFECT == _EDGEDETECT)
	shader.setUniform("kernelUnit", 1); // Add sampler uniform

	// Mirrored texture allows to prevent blackening caused by out-of-range accesses
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);

#define kernelX 3
#define kernelY 3

#if (_EFFECT == _SHARPEN)
	// Set texture with the sharpen kernel
#if 0
#define kernelX 5
#define kernelY 5
	float weights[25] = {-1.f, -1.f, -1.f, -1.f, -1.f,
						-1.f, -1.f, -1.f, -1.f, -1.f,
						-1.f, -1.f, 25.f, -1.f, -1.f,
						-1.f, -1.f, -1.f, -1.f, -1.f,
						-1.f, -1.f, -1.f, -1.f, -1.f};
#else
	float weights[kernelX*kernelY];
	for(int i = 0; i < kernelX*kernelY; i++)
		weights[i] = -1.f;
	weights[(kernelY/2)*kernelX+(kernelX/2)] = kernelX*kernelY;
#endif
	shader.setUniform("normalize", 1);

#elif (_EFFECT == _EDGEDETECT)
#if 0
	// Set texture with the Laplace kernel
	float weights[kernelX*kernelY];
	for(int i = 0; i < kernelX*kernelY; i++)
		weights[i] = -1.f;
	weights[(kernelY/2)*kernelX+(kernelX/2)] = kernelX*kernelY-1;
#else
	// Set texture with the Sobel kernel
#define kernelX 3
#define kernelY 3
	float weightsX[9] = {-1.f, 0.f, 1.f,
						-2.f, 0.f, 2.f,
						-1.f, 0.f, 1.f};
	float weightsY[9] = {-1.f, -2.f, -1.f,
						0.f, 0.f, 0.f,
						1.f, 2.f, 1.f};
	float *weights = weightsY;
#endif
	shader.setUniform("normalize", 0);
#endif

	// Send sizes to GLSL
	int kernelSize[2] = {kernelX, kernelY};
	shader.setUniform("kernelSize", kernelSize);
	shader.setUniform("grayscale", 0);

	// Kernel texture setup
	glActiveTexture(GL_TEXTURE1);
	kernelID = setKernelTexture(weights, kernelX, kernelY);
	glBindTexture(GL_TEXTURE_2D, kernelID);
#endif

	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

	return true;
}

void initVSL()
{
//	Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::PROJ_VIEW_MODEL, "m_pvm");
}


// ------------------------------------------------------------
//
// Main function
//

int main(int argc, char **argv)
{
//  GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA);

	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(512,512);
	glutCreateWindow("PV227 - Image Processing");


//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);

//	Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	int param;
	glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
	if(param == GL_CONTEXT_CORE_PROFILE_BIT)
		printf("Context Profile: Core\n");
	else
		printf("Context Profile: Compatibility\n");

	if(!setupShaders())
		exit(1);

	if(!init())
		exit(1);

	initVSL();

//  GLUT main loop
	glutMainLoop();

	return(0);
}
