#version 330

uniform sampler2D texUnit;

out vec4 outputF;

//----------------------------------------------------------------
// TODO:
// 1) Blur the image by averaging the 9 surrounding values (8-neighbourhood).
// 2) Protect against the out-of-range accesses (needs texture size - function "textureSize").
// 3) Blur the image by averaging the 25 surrounding values in the same way.
// 4) Blur the image with Gaussian blur kernel.
// 5) Notice the problems with the codes and the results.
//----------------------------------------------------------------

//----------------------------------------------------------------
// Problems:
// 1) Different codes for different sizes.
// 2) Blackening around the edges.
// 3) Need for various coeficients.
// 4) Processing colors separately may not be optimal - we will not deal with that :D
//----------------------------------------------------------------

void main()
{
	ivec2 texSize = textureSize(texUnit, 0);
	ivec2 pos = ivec2(gl_FragCoord);
	ivec2 ofs;
	vec4 sum = vec4(0.f);

#if 0
	for(ofs.y = -1; ofs.y <= 1; ofs.y++)
	{
		for(ofs.x = -1; ofs.x <= 1; ofs.x++)
		{
			// Not needed, maybe platform dependent
			if(any(lessThan(pos+ofs, ivec2(0)) || greaterThan(pos+ofs, texSize)))
				continue;

			vec4 texel = texelFetch(texUnit, pos+ofs, 0);
			sum += texel;
		}
	}
	
	outputF = sum/9.f;
#elif 0
	for(ofs.y = -2; ofs.y <= 2; ofs.y++)
	{
		for(ofs.x = -2; ofs.x <= 2; ofs.x++)
		{
			// Not needed, maybe platform dependent
			if(any(lessThan(pos+ofs, ivec2(0)) || greaterThan(pos+ofs, texSize)))
				continue;

			vec4 texel = texelFetch(texUnit, pos+ofs, 0);
			sum += texel;
		}
	}
	
	outputF = sum/25.f;
#else
	/*float weights[9] = float[9](1.f, 2.f, 1.f,
								2.f, 4.f, 2.f,
								1.f, 2.f, 1.f);*/

	float weights[25] = float[25](2.f, 7.f, 12.f, 7.f, 2.f,
								7.f, 31.f, 52.f, 31.f, 7.f,
								15.f, 52.f, 127.f, 52.f, 15.f,
								7.f, 31.f, 52.f, 31.f, 7.f,
								2.f, 7.f, 12.f, 7.f, 2.f);
	float sumWeight = 0.f;

	for(ofs.y = -2; ofs.y <= 2; ofs.y++)
	{
		for(ofs.x = -2; ofs.x <= 2; ofs.x++)
		{
			// Not needed, maybe platform dependent
			if(any(lessThan(pos+ofs, ivec2(0)) || greaterThan(pos+ofs, texSize)))
				continue;

			vec4 texel = texelFetch(texUnit, pos+ofs, 0);
			float weight = weights[(ofs.y+2)*5+(ofs.x+2)];
			sum += texel * weight;
			sumWeight += weight;
		}
	}
	
	outputF = sum/sumWeight;
#endif
}